#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import json
import os
import re
from rouge_score import rouge_scorer
from rouge_score import tokenize as rouge_tokenize
from tokenizer import tokenize, split_into_sentences


def rouge_tokenize_utf(text, stemmer):
    """A Unicode-friendly (and non-stemming) version of rouge_score's tokenization function."""
    text = text.lower()

    # Before calculating the ROUGE score, rouge_score's tokenizer removes all non-alphanumeric ([^a-z0-9]) ASCII
    # characters from the text. This includes Unicode characters, so "fjölmiðlar" would become "fj lmi lar". We replace
    # it with [^\w]+, which preserves all alphabetic Unicode characters.
    text = re.sub(r"[^\w-]+", " ", text)
    tokens = text.split()

    return tokens


# Override rouge_score's tokenizer with our Unicode-friendly version.
rouge_tokenize.tokenize = rouge_tokenize_utf


def generate_oracle(text, gold, max_tokens=None, max_sents=None, target_score='rouge2', target_metric='recall'):
    """Find the sentences which yield the highest possible ROUGE score given certain length constraints.

    :param text: An unsummarized news article, represented as a list of sentences.
    :param gold: A human-generated summary, represented as a list of sentences.
    :param max_tokens: Stop adding more sentences to the oracle summary once it has at least this many tokens.
    :param max_sents: Stop adding more sentences to the oracle summary once it has at least this many sentences.
    :param target_score: The ROUGE score which should be maximized (rouge1, rouge2, rougeL or rougeLsum).
    :param target_metric: The metric which should be maximized (precision, recall, fmeasure).
    :return: A list of labels, one for each sentence in the text (1 for sentences in the oracle summary, 0 for others).
    """
    scorer = rouge_scorer.RougeScorer([target_score])
    metrics = {'precision': 0, 'recall': 1, 'fmeasure': 2}
    target_metric = metrics[target_metric]

    summary = "\n".join([" ".join([t.txt for t in tokenize(s) if t.txt]) for s in gold])

    sentence_lengths = {pos: len(s.split()) for pos, s in enumerate(text)}
    remaining_sentences = set(range(len(text)))
    selected = set()

    best_score = 0
    oracle_length = 0

    while remaining_sentences:
        # Try adding each remaining sentence to the current summary, sort them so they're in the correct order
        # and see what the resulting ROUGE score would be. Pick whichever sentence increases the ROUGE score the most.
        candidates = {num: "\n".join([text[n] for n in sorted(selected | {num})]) for num in remaining_sentences}
        scores = {n: scorer.score(summary, c)[target_score][target_metric] for n, c in candidates.items()}

        num = max(candidates, key=lambda s: scores[s])
        max_score = scores[num]

        # Stop if the ROUGE score doesn't improve.
        if max_score <= best_score:
            break

        # Otherwise, add the best sentence to the oracle summary.
        best_score = max_score
        selected.add(num)
        remaining_sentences.remove(num)
        oracle_length += sentence_lengths[num]

        # Stop if length constraints are met or exceeded.
        if max_tokens is not None and oracle_length >= max_tokens:
            break

        if max_sents is not None and len(selected) == max_sents:
            break

    return [int(num in selected) for num in range(len(text))], best_score


def split(text):
    """Tokenize and sentence-segmentize a document, one line at a time.

    :param text: A string to be tokenized.
    :return: A list of tokenized sentences.
    """
    lines = [line.strip() for line in re.split(r'[\n\r]+', text.strip())]
    return [s.strip() for line in lines for s in split_into_sentences(line) if line]


def process_entry(entry):
    """Tokenize a news article and generate labels for each sentence using an oracle summary.

    :param entry: A news article from the IceSum corpus.
    :return: A dictionary containing the original text (tokenized), gold summary, labels and the oracle summary's score.
    """
    text_sentences = split(entry['text'])
    gold_sentences = entry['summary'].splitlines()
    labels, score = generate_oracle(text_sentences, gold_sentences, target_score='rouge2', max_tokens=100)

    output = {
        'text': text_sentences,
        'summary': entry['summary'].splitlines(),
        'labels': labels,
        'score': score
    }

    return output


def main():
    with open('icesum.json', encoding='utf-8') as f:
        icesum = json.load(f)
        icesum = {eid: entry for eid, entry in icesum.items()}

    # Use the splits that were used in the paper.
    with open('splits.json', encoding='utf-8') as f:
        splits = json.load(f)

    # Generate the oracle summaries
    oracle = {split_name: {eid: process_entry(icesum[eid]) for eid in docs} for split_name, docs in splits.items()}

    # Create training, validation and test files in the TransformerSum format
    script_path = os.path.realpath(__file__)
    root_dir = os.path.dirname(script_path)
    transformersum_dir = os.path.join(root_dir, 'transformersum')
    os.makedirs(transformersum_dir, exist_ok=True)

    data_path = {
        'train': os.path.join(transformersum_dir, 'train.icesum.json'),
        'valid': os.path.join(transformersum_dir, 'val.icesum.json'),
        'test': os.path.join(transformersum_dir, 'test.icesum.json'),
    }

    processed = {}

    for split_name, docs in oracle.items():
        for eid, entry in docs.items():
            src_sents = entry['text']
            src_tokens = [s.split() for s in src_sents]
            processed[eid] = {'src': src_tokens, 'labels': entry['labels']}

    # The examples in the test set also need the gold summary
    for eid, entry in oracle['test'].items():
        processed[eid]['tgt'] = "<q>".join(entry['summary'])

    # Write each file in the data subdirectory
    for split_name, docs in oracle.items():
        entries = [processed[eid] for eid in docs]
        output_path = data_path[split_name]
        with open(output_path, 'w', encoding='utf-8') as f:
            json.dump(entries, f, ensure_ascii=False, separators=(',', ':'))


if __name__ == '__main__':
    main()
